/**
 *  Sega Saturn USB dev cart firmware for use in ROM
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/
**/

#ifndef _ROM_CODE_EXEC_H_
#define _ROM_CODE_EXEC_H_


#include <stdio.h>


/**
 * Description : ROM code availability bitfield
 * Type        : byte (8 bits)
 * Location    : flash + 256KB-1
 * Bit0-6 : indicates if ROM code is installed or not.
 * Bit7   : must be set to 0 -> prevent 0xFF to be misinterpreted as "all ROM codes enabled".
**/
#define ROM_CHECK_OFFSET   0x0003FFFF
#define ROM_CHECK_ADDRESS (0x02000000 + ROM_CHECK_OFFSET)




/**
 * Bit0 : USB dev cart ROM code.
 * Folder       : usb_cartrom
 * Max size     : 1536 bytes
 * Description  : Execute USB dev cart I/O polling code.
 * Parameters   : none
 * Return value : -1 when ROM code couldn't be executed.
**/
#define ROMCODE_MSK_USB_DEV_CART 0x81
#define ROMCODE_VAL_USB_DEV_CART 0x01
#define ROM_USB_EXEC_ADDRESS  0x0203F800
#define rom_usb_dev_cart_io() (((*((unsigned char*)ROM_CHECK_ADDRESS)) & ROMCODE_MSK_USB_DEV_CART) == ROMCODE_VAL_USB_DEV_CART ? ((int (*)(void))ROM_USB_EXEC_ADDRESS)() : -1)



/**
 * Bit1 : loadbin ROM code.
 * See SCFirmWare/loadbin folder for further details.
**/


/**
 * Bit2 : Crappy debugger ROM code.
 * Folder       : dbg
 * Max size     : 480 bytes
 * Description  : Crappy debugger for use on custom yabause only.
 *                Uses only a few RAM, so that other ROM code can use it.
 * Parameter 1  : char* string         string to log.
 * Parameter 2  : unsigned long param  log parameters.
 *                            bit 0-7  : contents ID (message, file I/O, etc)
 *                            bit 8-10 : value size (1, 2, 4 bytes)
 *                            bit 12   : 1:log value in decimal 0:log in hexadecimal
 * Parameter 3  : unsigned long val    : value to log.
 * Return value : none
**/
#define ROMCODE_MSK_LOG 0x84
#define ROMCODE_VAL_LOG 0x04
#define ROM_LOG_EXEC_ADDRESS  0x0203F400
#define rom_crappy_debugger(_STRING_, _PARAM_, _VAL_) if(((*((unsigned char*)ROM_CHECK_ADDRESS)) & ROMCODE_MSK_LOG) == ROMCODE_VAL_LOG) { if(*((unsigned char*)ROM_LOG_EXEC_ADDRESS) != 0xFF) { ((void (*)(char*, unsigned long, unsigned long))ROM_LOG_EXEC_ADDRESS)((char*)_STRING_, (unsigned long)_PARAM_, (unsigned long)_VAL_); } }


/** Extra: string and integer logging functions. */
/* Log string to custom yabause. */
#define ROM_DBG(_STRING_) rom_crappy_debugger(_STRING_, 0x00000000, 0x00000000);
/* Log integer in decimal format to custom yabause. */
#define ROM_IDBG(_VAL_) rom_crappy_debugger(NULL, 0x00001000, _VAL_);
/* Log integer in hexadecimal format to custom yabause. */
#define ROM_H1DBG(_VAL_) rom_crappy_debugger(NULL, 0x00000100, _VAL_);
#define ROM_H2DBG(_VAL_) rom_crappy_debugger(NULL, 0x00000200, _VAL_);
#define ROM_H4DBG(_VAL_) rom_crappy_debugger(NULL, 0x00000400, _VAL_);




/**
 * Bit3 : SD card access library ROM code.
 * see sdcard/sdcard_public.h for further details.
**/



/**
 * Bit4 : Unused.
 * Bit5 : Unused.
 * Bit6 : Unused.
 * Bit7 : Must be set to zero.
**/


#endif /* _ROM_CODE_EXEC_H_ */


