/* 
 * This software is licensed under terms of the GNU GENERAL PUBLIC LICENSE
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This software  is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Source code is available at http://ppcenter.free.fr
 */


#ifndef SATCOM_SATCART_H
#define SATCOM_SATCART_H


//----------------------------------------------------------------------
// Saturn Development Cartridge definitions.
// http://ppcenter.free.fr/satcart/
// 
// Cartridge memory space is divided in the following 1MB blocks:
//  0x220x xxxx : RW 1MB flash memory (2 x 512KB flash memory chips used)
//  0x221x xxxx : RW USB module fifo data
//  0x222x xxxx : RW USB module fifo status register
//  0x223x xxxx : RW cartridge version, SD card related registers, etc
//  0x224x xxxx : RW 128KB SRAM
//  0x225x xxxx :    unused
//  0x226x xxxx :    unused
//  0x227x xxxx :    unused
//----------------------------------------------------------------------

#define CART_BASE_ADR 0x22000000


/******************************************
 * Flash memory address.
******************************************/
#define CART_FLASH (CART_BASE_ADR + 0)

/**
 * Maximum length allowed for firmware.
 * Can't use 1024KB, because not enough memory is available for FW flash.
 */
#define FIRM_MAXLEN (960*1024)


/******************************************
 * USB dev cart related definitions.
 * From antime's website : www.iki.fi/Anders.Montonen/sega/usbcart/
******************************************/
#define USB_FLAGS (*(volatile unsigned char*)(CART_BASE_ADR + 0x200001))
#define USB_RXF     (1 << 0)
#define USB_TXE     (1 << 1)
#define USB_PWREN   (1 << 7)
#define USB_FIFO (*(volatile unsigned char*)(CART_BASE_ADR + 0x100001))



/******************************************
 * CPLD related definitions.
******************************************/
#define CPLD_BASE_ADDR (CART_BASE_ADR + 0x300000)

/*
 * Offset : 0x00
 * Size   : 32bit
 * Access : read only
 * Description : CPLD build date
 * Read details :
 *  bit24-31: build year  (HEX)
 *  bit16-23: build month (HEX)
 *  bit8-15 : build day   (HEX)
 *  bit0-7  : build hour  (HEX)
 */
#define REG_CPLD_DATE (*(volatile unsigned long*)(CPLD_BASE_ADDR + 0x00))

/*
 * Offset : 0x04
 * Size   : 16bit
 * Access : read only
 * Description : CPLD/target PCB version
 * Read details :
 *  bit0-3   : CPLD minor version (hexadecimal format)
 *  bit4-7   : CPLD major version (hexadecimal format)
 *  bit8-15  : Target PCB version
 */
#define REG_CART_CPLD_VER (*(volatile unsigned short*)(CPLD_BASE_ADDR + 0x04))

/*
 * Offset : 0x06
 * Size   : 16bit
 * Access : read only
 * Description : Functionalities list register
 * Read details :
 *  bit0    : USB module              (0:unavailable, 1:available)
 *  bit1    : SPI bitbang SD card     (0:unavailable, 1:available)
 *  bit2    : Serial 4bits SD card    (0:unavailable, 1:available)
 *  bit3    : Reserved. Returns zero.
 *  bit4    : 128KB SRAM chip         (0:unavailable, 1:available)
 *  bit5~15 : Reserved. Returns zero.
 */
#define REG_CPLD_FUNC (*(volatile unsigned short*)(CPLD_BASE_ADDR + 0x06))
#define CPLD_FUNC_USB_RSHFT     0
#define CPLD_FUNC_SDBIT_RSHFT   1
#define CPLD_FUNC_SDSER4_RSHFT  2
#define CPLD_FUNC_SRAM128_RSHFT 4

/*
 * Offset : 0x0C
 * Size   : 16bit
 * Access : read only
 * Description : Return 0x5555 on data bus
 */
#define REG_CPLD_5555 (*(volatile unsigned short*)(CPLD_BASE_ADDR + 0x0C))
/*
 * Offset : 0x0E
 * Size   : 16bit
 * Access : read only
 * Description : Return 0xAAAA on data bus
 */
#define REG_CPLD_AAAA (*(volatile unsigned short*)(CPLD_BASE_ADDR + 0x0E))




/******************************************
 * SD card related definitions.
******************************************/
#define SD_BASE_ADDR (CART_BASE_ADR + 0x300010)







/******************************************
 * SRAM (128KB) related definitions.
 * All sizes are byte unit.
 *
 * Memory regions :
 *  For satcom_lib
 *   + global data
 *   + internal data
 *   + debugger data
 *  For sd card rom code
 *   + various data
 *   + sector cache data
 *  Free for user
******************************************/
#define CART_SRAM_ADR (CART_BASE_ADR + 0x400000)
#define CART_SRAM_SIZ (128*1024)

/* SRAM top. */
#define CART_SRAM_PTR ((volatile unsigned char*)(CART_SRAM_ADR))

/* SatCom area (8KB). */
#define CART_SRAM_SATCOM_PTR (CART_SRAM_PTR + 0)
#define CART_SRAM_SATCOM_SIZ (8 * 1024)
/* Global data start from the beginning of SatCom area. */
#define SC_GLOBAL_DATA_RAM_ADR (CART_SRAM_ADR)


/* SD card ROM code area (56KB). */
#define CART_SRAM_SDRC_PTR (CART_SRAM_PTR + CART_SRAM_SATCOM_SIZ)
#define CART_SRAM_SDRC_SIZ (56 * 1024)


/* User area (remaining until 128KB). */
#define CART_SRAM_USER_PTR (CART_SRAM_PTR + CART_SRAM_SATCOM_SIZ + CART_SRAM_SDRC_SIZ)
#define CART_SRAM_USER_SIZ (CART_SRAM_SIZ - CART_SRAM_SDRC_SIZ - CART_SRAM_SATCOM_SIZ)



#endif // SATCOM_SATCART_H

