/* This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details. */ 
/* Latest source code is available at http://ppcenter.free.fr. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>


/* IP.BIN header format. */
#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif

typedef struct _ip_bin
{
    /* Text-based informations. */
    char hwr_id[16]; // "SEGA SEGASATURN "
    char man_id[16]; // "SEGA TP KAISHA-A"
    char prd_nb[10]; // "T-053G    "
    char version[6]; // "V0.001"
    char date[8];    // "19940912"
    char dev_id[8];  // "CD-1/1  "
    char area[16];   // "JTUBKAEL        "
    char pers[16];   // "JAST            "
    char title[112]; // ...

    /* Value-based informations. */
    unsigned char unused01[4];
    unsigned char unused02[4];
    unsigned char unused03[4];
    unsigned char unused04[4];

    unsigned char ip_size[4];   // 0x00001800
    unsigned char unused06[4];
    unsigned char stack_m[4];   // 0x00000000 ?
    unsigned char stack_s[4];   // 0x00000000 ?

    unsigned char first_adr[4]; // 0x06004000
    unsigned char first_siz[4]; // 0x00000000 ?
    unsigned char unused11[4];
    unsigned char unused12[4];
} ip_bin_t;

#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif


unsigned char char2nchar(unsigned char c)
{
    if(c < '0') return '0';
    if(c > '9') return '9';
    return c;
}




int main (int argc, char *argv[])
{
    printf("--- IP header customizer v0.3 by cafe-alpha. ---\n");
    int verbose = 0;
    int offset  = 0;

    FILE* fp = NULL;
    /* Our IP header. */
    ip_bin_t header;
    memset((void*)(&header), ' ', sizeof(ip_bin_t));

    if(argc <= 1)
    {
        printf( "Type %s -h for help.\n", argv[0]);
        return 0;
    }
    if(argc >= 2)
    {
        if((!strcmp(argv[1], "-h"))
        || (!strcmp(argv[1], "-help"))
        || (!strcmp(argv[1], "-?")))
        {
            printf(" Usage : <in_file> [-xxx [<str>]] [-verbose] ...\n");
            printf("  <in_file>  can be an ISO, or BIN/CUE image, or even Cartridge firmware file.\n");
            printf("  [-verbose] be verbose.\n");
            printf("  IP header  customization arguments:\n");
            printf("   -man        <str> : set ip manufacturer    (ex:SEGA TP KAISHA-A)\n");
            printf("   -prod       <str> : set ip product number  (ex:T-053G)\n");
            printf("   -ver        <str> : set ip version         (ex:V0.001)\n");
            printf("   -autover    <file>: set ip version same as in release.h\n");
            printf("   -date       <str> : set ip date            (ex:19940912)\n");
            printf("   -today            : set ip date to today\n");
            printf("   -dev        <str> : set ip device ID       (ex:CD-1/1)\n");
            printf("   -area       <str> : set ip area            (ex:JTUBKAEL)\n");
            printf("   -per        <str> : set ip peripherals     (ex:JAST)\n");
            printf("   -title      <str> : set ip title           (ex:MY TITLE)\n");
            printf(" \n");
            printf("   -ip_size    <val> : set ip size            (ex:0x00001800)\n");
            printf("   -stack_m    <val> : set CPU1 stack address (ex:0x06001000)\n");
            printf("   -stack_s    <val> : set CPU2 stack address (ex:0x06000D00)\n");
            printf("   -first_adr  <val> : set 1st read address   (ex:0x06004000)\n");
            printf("   -first_siz  <val> : set 1st read size      (ex:0x00000000)\n");
            return 0;
        }
    }

    /* Open file, and read its header. */
    printf(" Open file: %s\n", argv[1]);
    fp = fopen(argv[1], "rb");
    if(fp == NULL)
    {
        printf(" IP header: couldn't open file for reading !\n");
        return 0;
    }
    fread((void*)&header, sizeof(ip_bin_t), 1, fp);

    /* BIN/CUE format : header starts after 16 bytes of data. */
    char* raw_data = (char*)&header;
    if(memcmp((void*)(raw_data+16), (void*)"SEGA SEGASATURN ", 16) == 0)
    {
        offset = 16;
        /* Re-read header from new offset. */
        fseek(fp, offset, SEEK_SET);
        fread((void*)&header, sizeof(ip_bin_t), 1, fp);
    }

    fclose(fp); fp = NULL;


    int i;

    // Parse arguments
    for(i=2; i<argc; i++)
    {
        if(!strcmp(argv[i], "-man"))
        {i++;
            memset((void*)(&(header.man_id)), ' ', sizeof(header.man_id));
            strncpy(header.man_id, argv[i], sizeof(header.man_id));
        }
        else if(!strcmp(argv[i], "-prod"))
        {i++;
            memset((void*)(&(header.prd_nb)), ' ', sizeof(header.prd_nb));
            strncpy(header.prd_nb, argv[i], sizeof(header.prd_nb));
        }
        else if(!strcmp(argv[i], "-ver"))
        {i++;
            memset((void*)(&(header.version)), ' ', sizeof(header.version));
            strncpy(header.version, argv[i], sizeof(header.version));
        }
        else if(!strcmp(argv[i], "-autover"))
        {i++;
            unsigned long len;
            unsigned char version_data[8]; //V1.234\r\n

            /* Print dummy version in the case autoversion failed. */
            strncpy(header.version, "V1.234", sizeof(header.version));
            strncpy((char*)version_data, header.version, sizeof(version_data));

            FILE* fpr = fopen(argv[i], "rb");
            if(fpr)
            {
                fseek(fpr, 0, SEEK_END);
                len = ftell(fpr);
                if(len >= sizeof(version_data))
                {
                    len -= sizeof(version_data);
                    fseek(fpr, len, SEEK_SET);
                    fread((void*)version_data, 1, sizeof(version_data), fpr);
                    version_data[0] = 'V';
                    version_data[1] = char2nchar(version_data[1]);
                    version_data[2] = '.';
                    version_data[3] = char2nchar(version_data[3]);
                    version_data[4] = char2nchar(version_data[4]);
                    version_data[5] = char2nchar(version_data[5]);
                    version_data[6] = '\0';
                    version_data[7] = '\0';
                    strncpy(header.version, (char*)version_data, sizeof(header.version));
                }
                fclose(fpr);
            }
            else
            {
                printf(" Open %s file error !\n", argv[i]);
            }
        }
        else if(!strcmp(argv[i], "-date"))
        {i++;
            memset((void*)(&(header.date)), ' ', sizeof(header.date));
            strncpy(header.date, argv[i], sizeof(header.date));
        }
        else if(!strcmp(argv[i], "-today"))
        {
            time_t     now;
            struct tm  *ts;
            now = time(NULL);
            ts = localtime(&now);
            char       buf[80];
            strftime(buf, sizeof(buf), "%Y%m%d", ts);
            strncpy(header.date, buf, sizeof(header.date));
        }
        else if(!strcmp(argv[i], "-dev"))
        {i++;
            memset((void*)(&(header.dev_id)), ' ', sizeof(header.dev_id));
            strncpy(header.dev_id, argv[i], sizeof(header.dev_id));
        }
        else if(!strcmp(argv[i], "-area"))
        {i++;
            memset((void*)(&(header.area)), ' ', sizeof(header.area));
            strncpy(header.area, argv[i], sizeof(header.area));
        }
        else if(!strcmp(argv[i], "-per"))
        {i++;
            memset((void*)(&(header.pers)), ' ', sizeof(header.pers));
            strncpy(header.pers, argv[i], sizeof(header.pers));
        }
        else if(!strcmp(argv[i], "-title"))
        {i++;
            memset((void*)(&(header.title)), ' ', sizeof(header.title));
            strncpy(header.title, argv[i], sizeof(header.title));
        }
        else if(!strcmp(argv[i], "-ip_size"))
        {i++;
            unsigned long ip_size = strtoul(argv[i], NULL, 0);
            printf(" ip_size = %X", ip_size >> 24);
            header.ip_size[0] = ( ip_size >> 24 ) & 0xFF;
            printf(" ip_size test = %X", header.ip_size[0]);
            header.ip_size[1] = ( ip_size >> 16 ) & 0xFF;
            header.ip_size[2] = ( ip_size >>  8 ) & 0xFF;
            header.ip_size[3] = ( ip_size >>  0 ) & 0xFF;
        }
        else if(!strcmp(argv[i], "-stack_m"))
        {i++;
            unsigned long stack_m = strtoul(argv[i], NULL, 0);
            header.stack_m[0] = ( stack_m >> 24 ) & 0xFF;
            header.stack_m[1] = ( stack_m >> 16 ) & 0xFF;
            header.stack_m[2] = ( stack_m >>  8 ) & 0xFF;
            header.stack_m[3] = ( stack_m >>  0 ) & 0xFF;
        }
        else if(!strcmp(argv[i], "-stack_s"))
        {i++;
            unsigned long stack_s = strtoul(argv[i], NULL, 0);
            header.stack_s[0] = ( stack_s >> 24 ) & 0xFF;
            header.stack_s[1] = ( stack_s >> 16 ) & 0xFF;
            header.stack_s[2] = ( stack_s >>  8 ) & 0xFF;
            header.stack_s[3] = ( stack_s >>  0 ) & 0xFF;
        }
        else if(!strcmp(argv[i], "-first_adr"))
        {i++;
            unsigned long first_adr = strtoul(argv[i], NULL, 0);
            header.first_adr[0] = ( first_adr >> 24 ) & 0xFF;
            header.first_adr[1] = ( first_adr >> 16 ) & 0xFF;
            header.first_adr[2] = ( first_adr >>  8 ) & 0xFF;
            header.first_adr[3] = ( first_adr >>  0 ) & 0xFF;
        }
        else if(!strcmp(argv[i], "-first_siz"))
        {i++;
            unsigned long first_siz = strtoul(argv[i], NULL, 0);
            header.first_siz[0] = ( first_siz >> 24 ) & 0xFF;
            header.first_siz[1] = ( first_siz >> 16 ) & 0xFF;
            header.first_siz[2] = ( first_siz >>  8 ) & 0xFF;
            header.first_siz[3] = ( first_siz >>  0 ) & 0xFF;
        }
        else if(!strcmp(argv[i], "-verbose"))
        {
            verbose = 1;
        }
        else
        {
            printf(" IP header: unknown parameter ! (%s)\n", argv[i]);
        }
    }

    /* Remove non-ASCII characters in text-based informations. */
    unsigned char* str = (unsigned char*)&header;
    for(i-0; i<208; i++)
    {
        if(str[i] < ' ')
        {
            str[i] = ' ';
        }
    }


    /* If requested, display header contents. */
    if(verbose)
    {
        char str_tmp[200];
        int l;
        #define DISP_FIELD(_F_) l=7-strlen(#_F_); memset(str_tmp, ' ', sizeof(str_tmp)); memset(str_tmp+l, 0, sizeof(str_tmp)-l); strncpy(str_tmp+l, header._F_, sizeof(header._F_)); printf("  %s=%s\n", #_F_, str_tmp);
        DISP_FIELD(hwr_id);
        DISP_FIELD(man_id);
        DISP_FIELD(prd_nb);
        DISP_FIELD(version);
        DISP_FIELD(date);
        DISP_FIELD(dev_id);
        DISP_FIELD(area);
        DISP_FIELD(pers);
        DISP_FIELD(title);
    }


    /* Write modified header to file. */
    fp = fopen(argv[1], "r+");
    if(fp == NULL)
    {
        printf(" IP header: couldn't open file for writing !\n");
        return 0;
    }
    fseek(fp, offset, SEEK_SET);
    fwrite((void*)&header, sizeof(ip_bin_t), 1, fp);
    fclose(fp); fp = NULL;

    return 0;
}



