/* This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details. */ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../../satcom_lib/sc_common.h"
#include "../../satcom_lib/saturn/rom_code/rom_code_header.h"
#include "../../satcom_lib/saturn/rom_code/rom_code_exec.h"

/* Binary data of IP.BIN and loader. */
#include "data.h"
//#include "data.c"

/* Offset where to put the IP.BIN / loader / compressed data. */
#include "mf_address.h"

/* Maximum firmware length. */
#include "../../satcom_lib/saturn/satcart.h"

/* Endianness conversion. */
unsigned long bswap_ul(unsigned long in)
{
    unsigned long out;
    out  = ((in >> 24) & 0xFF) <<  0;
    out |= ((in >> 16) & 0xFF) <<  8;
    out |= ((in >>  8) & 0xFF) << 16;
    out |= ((in >>  0) & 0xFF) << 24;
    return out;
}

#define EXTRA_BIN_COUNT 8


#define KB_I(_VAL_) ((_VAL_)>>10)
#define KB_F(_VAL_) ((((_VAL_)%1024) * 1000)/1024)


int main (int argc, char *argv[])
{
    int verbose = 1;

    int i;
    if(argc <= 1)
    {
        printf(" Type %s -h for help.\n", argv[0]);
        return 0;
    }

    if((!strcmp(argv[1], "-rom_code_patch_silent")) || (!strcmp(argv[1], "-rom_code_silent")))
    {
        verbose = 0;
    }
    if(verbose)
    {
        printf("--- Saturn FirmWare maker v0.4 by cafe-alpha. ---\n");
    }

    FILE* fp = NULL;

    char* infile = argv[1];
    char* outfile = argv[2];

    unsigned long exec_addr = 0x06004000;

    unsigned long rom_size = 256;
    unsigned long exec_size = 256;

    int romcode_footer_use = 0;
    int extra_bin_count = 0;
    char* extra_bin_file[EXTRA_BIN_COUNT];
    unsigned long extra_bin_offset[EXTRA_BIN_COUNT];
    unsigned char footer_romcode = 0;


    if(argc >= 2)
    {
        if((!strcmp(argv[1], "-h"))
        || (!strcmp(argv[1], "-help"))
        || (!strcmp(argv[1], "-?")))
        {
            printf(" Usage 1 : <in_file> <out_file> [-silent]\n");
            printf("                               [-address <exec_address>]\n");
            printf("                               [-rom_size <rom_size_KB>]\n");
            printf("                               [-exec_max_size <exec_max_size_KB>]\n");
            printf("                               [-extra_bin <file.bin> <rom_offset> <romcode_value>]\n");
            printf("  Default values:\n");
            printf("          exec_address     : 0x%08X\n", exec_addr);
            printf("          rom_size_KB      : %d\n", rom_size);
            printf("          exec_max_size_KB : %d\n", exec_size);
            printf("          footer           : Don't use\n");
            printf("          extra_bin        : Don't use\n");
            printf("\n");
            printf(" Usage 2 : -rom_code <in_file> <out_file>\n");
            printf("                     <version_major> <version_minor>\n");
            printf("                     <addr> <maxlen>\n");
            printf("                     <version_string1> <version_string2>\n");
            printf("                     <romcode_id>\n");
            printf("                     <compatible_hw_id>\n");
            printf("\n");
            printf(" Usage 3 : -rom_code_patch <in_file> <rc_file>\n");
            return 0;
        }
    }


    /* Make ROM code ? */
    if((!strcmp(argv[1], "-rom_code")) || (!strcmp(argv[1], "-rom_code_silent")))
    {
        sc_rom_code_header_t h;
        infile = argv[2];
        outfile = argv[3];
        unsigned char version_major = strtoul(argv[4], NULL, 0);
        unsigned char version_minor = strtoul(argv[5], NULL, 0);
        unsigned long write_address = strtoul(argv[6], NULL, 0);
        unsigned long write_maxlen  = strtoul(argv[7], NULL, 0);
        char* version_string1 = argv[8];
        char* version_string2 = argv[9];
        unsigned char romcode_id = strtoul(argv[10], NULL, 0);
        unsigned char compatible_hw_id = strtoul(argv[11], NULL, 0);

        memset((void*)&h, 0x00, sizeof(sc_rom_code_header_t));

        h.magic_42 = 0x42;
        h.magic_53 = 0x53;

        h.version_major = version_major;
        h.version_minor = version_minor;

        h.write_address = bswap_ul(write_address);
        h.write_maxlen  = bswap_ul(write_maxlen);

        h.romcode = romcode_id;
        h.compatible_hw_id = compatible_hw_id;

        strncpy(h.version_string1, version_string1, sizeof(h.version_string1));
        h.version_string1[sizeof(h.version_string1) - 1] = '\0';

        strncpy(h.version_string2, version_string2, sizeof(h.version_string2));
        h.version_string2[sizeof(h.version_string2) - 1] = '\0';

        printf("Build ROM code v. %02X.%02X \"%s\" \"%s\". Wadr=0x%08X mlen=%d rc=0x%02X hw=0x%02X ...\n", 
            h.version_major, h.version_minor, 
            h.version_string1, h.version_string2, 
            write_address, write_maxlen, 
            h.romcode, 
            h.compatible_hw_id);


        unsigned long inlen;
        fp = fopen(infile, "rb");
        if(!fp)
        {
            printf(" mkfirm error : open input file error !\n");
            return 1;
        }
        fseek(fp, 0, SEEK_END);
        inlen = ftell(fp);
        fseek(fp, 0, SEEK_SET);

        if((sizeof(sc_rom_code_header_t) + inlen) > write_maxlen)
        {
            fclose(fp); fp = NULL;
            printf(" mkfirm error : ROM code too large !\n");
            return 1;
        }

        unsigned char* inbuffer = (unsigned char*)malloc(inlen);
        if(!inbuffer)
        {
            fclose(fp); fp = NULL;
            printf(" mkfirm error : memory allocation !\n");
            return 1;
        }
        fread(inbuffer, 1, inlen, fp);
        fclose(fp); fp = NULL;

        fp = fopen(outfile, "wb");
        if(fp)
        {

            fwrite((void*)&h, 1, sizeof(sc_rom_code_header_t), fp);
            fwrite(inbuffer, 1, inlen, fp);

            if(verbose)
            {
                printf(" Wrote %d bytes to disk, remaining space = %d.%03d KB\n", 
                    sizeof(sc_rom_code_header_t) + inlen, 
                    KB_I(write_maxlen - (sizeof(sc_rom_code_header_t) + inlen)), 
                    KB_F(write_maxlen - (sizeof(sc_rom_code_header_t) + inlen)));
            }
            fclose(fp); fp = NULL;
        }
        else
        {
            printf(" Error : can't open file (%s) !\n", outfile);
        }

        free(inbuffer); inbuffer = NULL;
        return 0;
    }

    /* Patch ROM code ? */
    if((!strcmp(argv[1], "-rom_code_patch")) || (!strcmp(argv[1], "-rom_code_patch_silent")))
    {
        sc_rom_code_header_t h;

        infile = argv[2];
        char* rcfile = argv[3];
        unsigned long fwlen = 0;
        unsigned long inlen = 0;
        unsigned long outlen = FIRM_MAXLEN;
        unsigned char* inbuffer = (unsigned char*)malloc(outlen);
        if(!inbuffer)
        {
            printf(" mkfirm error : memory allocation !\n");
            return 1;
        }
        memset((void*)inbuffer, 0xFF, outlen);

        printf("Patch firmware (%s) with ROM code (%s) ...\n", infile, rcfile);

        /* Open Firmware. */
        fp = fopen(infile, "rb");
        if(!fp)
        {
            free(inbuffer); inbuffer = NULL;
            printf(" mkfirm error : open input file error !\n");
            return 1;
        }
        fseek(fp, 0, SEEK_END);
        inlen = ftell(fp); fwlen = inlen;
        fseek(fp, 0, SEEK_SET);

        fread(inbuffer, 1, inlen, fp);
        fclose(fp); fp = NULL;

        /* Open ROM code file. */
        fp = fopen(rcfile, "rb");
        if(!fp)
        {
            free(inbuffer); inbuffer = NULL;
            printf(" mkfirm error : open ROM code file error !\n");
            return 1;
        }
        fread((void*)&h, 1, sizeof(sc_rom_code_header_t), fp);
        fseek(fp, 0, SEEK_END);
        inlen = ftell(fp);
        fseek(fp, 0, SEEK_SET);

        /* Copy ROM code. */
        if(verbose)
        {
            printf(" write_address = 0x%08X\n", bswap_ul(h.write_address));
        }
        fread(inbuffer + (bswap_ul(h.write_address) & 0x00FFFFFF), 1, inlen, fp);
        fclose(fp); fp = NULL;

        /* Patch ROM code check bit. */
        unsigned char* romcode_bf = (unsigned char*)(inbuffer + ROM_CHECK_OFFSET);
        //printf(" romcode_bf = 0x%02X\n", *romcode_bf);
        if((*romcode_bf) & 0x80)
        { /* Check bit init. */
            (*romcode_bf) = 0x00;
        }
        (*romcode_bf) |= h.romcode;
        //printf(" romcode_bf = 0x%02X\n", *romcode_bf);



        /* Write patched stuff to disk. */
        if(((bswap_ul(h.write_address)&0x00FFFFFF) + bswap_ul(h.write_maxlen)) <= fwlen)
        {
            outlen = fwlen;
        }
        if(verbose)
        {
            printf(" fwlen = %d, inlen = %d, outlen = %d rc_h=%d\n", fwlen, inlen, outlen, sizeof(sc_rom_code_header_t));
        }
        fp = fopen(infile, "wb");
        if(fp)
        {
            fwrite(inbuffer, 1, outlen, fp);

            if(verbose)
            {
                printf(" Wrote %d bytes to disk.\n", outlen);
            }
            fclose(fp); fp = NULL;
        }
        else
        {
            printf(" Error : can't open file (%s) !\n", infile);
        }

        free(inbuffer); inbuffer = NULL;
        return 0;
    }



    // Parse arguments
    for(i=3; i<argc; i++)
    {
        if(!strcmp(argv[i], "-silent"))
        {
            verbose = 0;
        }
        else if(!strcmp(argv[i], "-address"))
        {i++;
            exec_addr = strtoul(argv[i], NULL, 0);
        }
        else if(!strcmp(argv[i], "-rom_size"))
        {i++;
            rom_size = strtoul(argv[i], NULL, 0);
            if(rom_size < 64) rom_size = 64;
        }
        else if(!strcmp(argv[i], "-exec_max_size"))
        {i++;
            exec_size = strtoul(argv[i], NULL, 0);
        }
        else if(!strcmp(argv[i], "-extra_bin"))
        {i++;
            unsigned char rc;
            extra_bin_file[extra_bin_count]   = argv[i+0];
            extra_bin_offset[extra_bin_count] = strtoul(argv[i+1], NULL, 0);
            extra_bin_count++;

            /* Update footer's ROM code value ? */
            rc = strtoul(argv[i+2], NULL, 0);
            if(rc != 0)
            {
                footer_romcode |= rc;
                romcode_footer_use = 1;
            }

            i+=2;
        }
        else
        {
            printf(" mkfirm error : unknown parameter ! (%s)\n", argv[i]);
        }
    }


    if(verbose)
    {
        printf(" Input file   : %s\n", infile);
        printf(" Output file  : %s\n", outfile);
        printf(" Exec address : 0x%08X\n", exec_addr);
        printf(" BIN size (KB): %d\n", exec_size);
        printf(" ROM size (KB): %d\n", rom_size);
        for(i=0; i<extra_bin_count; i++)
        {
            printf(" Extra binary#%d : file=%s offset=0x%08X\n", i, extra_bin_file[i], extra_bin_offset[i]);
        }
    }
    rom_size = rom_size*1024;
    exec_size = exec_size*1024;

    /* Keep previous footer data. */
    unsigned char prevromcode_bf = 0xFF;

    fp = fopen(outfile, "rb");
    if(fp)
    {
        fseek(fp, ROM_CHECK_OFFSET, SEEK_SET);
        fread((void*)&prevromcode_bf, 1, sizeof(prevromcode_bf), fp);
        fclose(fp); fp = NULL;
    }


    /* Delete output file. */
    unlink(outfile);

    /* Read target program file. */
    unsigned long inlen;
    fp = fopen(infile, "rb");
    if(!fp)
    {
        printf(" mkfirm error : open input file error !\n");
        return 1;
    }
    fseek(fp, 0, SEEK_END);
    inlen = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    unsigned char* inbuffer = (unsigned char*)malloc(inlen);
    if(!inbuffer)
    {
        fclose(fp); fp = NULL;
        printf(" mkfirm error : memory allocation !\n");
        return 1;
    }
    fread(inbuffer, 1, inlen, fp);
    fclose(fp); fp = NULL;



    /* Compress target program. */
    if(verbose) printf(" Compress data ...\n");
    unsigned char* compbuffer = (unsigned char*)malloc(inlen+SC_COMP_HEADER_MAXSIZE);
    if(!compbuffer)
    {
        free(inbuffer); inbuffer = NULL;
        printf(" mkfirm error : memory allocation !\n");
        return 1;
    }
    unsigned long complen;
    // Compress user program.
    complen = comp_exec(inbuffer, inlen, compbuffer);
    if(verbose)
    {
        printf(" Input size      = %d bytes\n", inlen);
        printf(" Compressed size = %d bytes (input size * %f %%)\n", complen, ((float)(100*complen))/inlen);
    }
    free(inbuffer); inbuffer = NULL;

    /* Compressed file size check */
    if((FC_DATA_OFFSET + complen) >= (exec_size-256))
    {
        free(compbuffer); compbuffer = NULL;
        printf(" mkfirm error : data don't fit in given size !\n");
        return 1;
    }

    if(verbose)
    {
        printf(" Output file : alloc %d bytes ...\n", rom_size);
    }

    unsigned char* outbuffer = NULL;
    outbuffer = (unsigned char*)malloc(rom_size);
    if(!outbuffer)
    {
        printf(" mkfirm error : memory allocation !\n");
        return 1;
    }

    if(verbose)
    {
        printf(" Output file : reset firmware data ...\n");
    }

    memset((void*)outbuffer, 0xFF, rom_size);
    memset((void*)outbuffer, 'P', 64*1024); // For debug purpose.

    /* Copy ip.bin */
    if(verbose)
    {
        printf(" Output file : copy IP.BIN ...\n");
    }
    memcpy((void*)(outbuffer + FC_IPBIN_OFFSET), (void*)ip_bin_dat, IP_BIN_SIZ);

    /* Copy loader */
    if(verbose)
    {
        printf(" Output file : copy loader ...\n");
    }
    memcpy((void*)(outbuffer + FC_LOADER_OFFSET), (void*)loader_bin_dat, LOADER_BIN_SIZ);

    /* Copy decompressor */
    if(verbose)
    {
        printf(" Output file : copy decompressor ...\n");
    }
    memcpy((void*)(outbuffer + FC_DECOMPRESSOR_OFFSET), (void*)decompressor_bin_dat, DECOMPRESSOR_BIN_SIZ);



    /* Write loader parameters. */
    if(verbose)
    {
        printf(" Output file : write loader parameters ...\n");
    }
    /* Pointer to the parameters (length, execution start address). */
    unsigned long* params = (unsigned long*)(outbuffer + FC_PARAMS_OFFSET);
    params[FC_PDATALEN_INDEX] = bswap_ul(complen);
    params[FC_EXEADDRS_INDEX] = bswap_ul(exec_addr);

    /* Copy target program. */
    if(verbose)
    {
        printf(" Output file : copy target program ...\n");
    }
    memcpy((void*)(outbuffer + FC_DATA_OFFSET), (void*)compbuffer, complen);

    /* If needed, add ROM code footer data. */
    if(rom_size < (256*1024)) romcode_footer_use = 0;
    if(romcode_footer_use)
    {
        if(verbose)
        {
            printf(" Output file : add footer data ...\n");
        }
        memcpy((void*)(outbuffer + ROM_CHECK_OFFSET), (void*)&prevromcode_bf, sizeof(prevromcode_bf));
        unsigned char* romcode_bf = (unsigned char*)(outbuffer + ROM_CHECK_OFFSET);
        *romcode_bf = footer_romcode;
    }

    /* Add extra binary data. */
    for(i=0; i<extra_bin_count; i++)
    {
        if(verbose)
        {
            printf(" Output file : Add extra binary (%s) ...\n", extra_bin_file[i]);
        }

        FILE* fpe=fopen(extra_bin_file[i], "rb");
        if(fpe)
        {
            fseek(fpe, 0, SEEK_END);
            unsigned long extra_len = ftell(fpe);
            fseek(fpe, 0, SEEK_SET);

            if(extra_bin_offset[i] + extra_len > rom_size)
            {
                extra_len = rom_size - extra_bin_offset[i];
            }

            fread(outbuffer+extra_bin_offset[i], 1, extra_len, fpe);
            fclose(fpe); fpe = NULL;
        }
        else
        {
            printf(" mkfirm error : couldn't open extra binary (%s) !\n", extra_bin_file[i]);
        }
    }


    /* Write IP.BIN + loader + compressed data to disk. */
    if(verbose)
    {
        printf(" Output file : write data to disk ...\n");
    }
    fp = fopen(outfile, "wb");
    if(fp)
    {
        unsigned long writelen = rom_size;

        fwrite(outbuffer, 1, writelen, fp);
        printf(" Wrote %d.%03d KB to disk (Input = %d.%03d KB, comp. = %d.%03d KB -> input * %3.1f %%, remain %d.%03d KB)\n", 
            KB_I(writelen), KB_F(writelen), 
            KB_I(inlen), KB_F(inlen), 
            KB_I(complen), KB_F(complen), 
            ((float)(100*complen))/inlen, 
            KB_I(exec_size-complen), KB_F(exec_size-complen));
        fclose(fp); fp = NULL;
    }
    else
    {
        printf(" Error : can't open file (%s) !\n", outfile);
    }
    free(outbuffer); outbuffer = NULL;
    free(compbuffer); compbuffer = NULL;

    return 0;
}



